@@ -4,13 +4,14 @@ use strict;
 use warnings;
 use feature 'state';
 
-use Coro;
-use Coro::AnyEvent;
+use Fcntl qw(:flock);
 use File::Find;
 use File::Path;
 use File::stat;
 use File::Spec;
 use FindBin;
+use IO::Async::Loop;
+use IO::Async::Function;
 use List::Util qw(all any);
 use Path::Tiny;
 use Time::Piece;
@@ -20,8 +21,6 @@ use PLS::Trie;
 
 use constant {INDEX_LOCATION => File::Spec->catfile('.pls_cache', 'index')};
 
-my $indexing_semaphore = Coro::Semaphore->new();
-
 =head1 NAME
 
 PLS::Parser::Index
@@ -39,17 +38,19 @@ sub new
 
     my %args = @args;
 
-    my %self = (
-                root                     => $args{root},
-                location                 => File::Spec->catfile($args{root}, INDEX_LOCATION),
-                cache                    => {},
-                subs_trie                => PLS::Trie->new(),
-                packages_trie            => PLS::Trie->new(),
-                last_mtime               => 0,
-                retrieve_index_semaphore => Coro::Semaphore->new()
-               );
+    my $self = bless {
+                      root          => $args{root},
+                      location      => File::Spec->catfile($args{root}, INDEX_LOCATION),
+                      cache         => {},
+                      subs_trie     => PLS::Trie->new(),
+                      packages_trie => PLS::Trie->new(),
+                      last_mtime    => 0,
+                     }, $class;
 
-    return bless \%self, $class;
+    $self->start_indexing_function();
+    $self->start_cleanup_function();
+
+    return $self;
 } ## end sub new
 
 sub load_trie
@@ -61,70 +62,131 @@ sub load_trie
     foreach my $sub (keys %{$index->{subs}})
     {
         my $count = scalar @{$index->{subs}{$sub}};
-        $self->{subs_trie}->insert($sub, 1);
+        $self->{subs_trie}->insert($sub, $count);
     }
 
     foreach my $package (keys %{$index->{packages}})
     {
         my $count = scalar @{$index->{packages}{$package}};
-        $self->{packages_trie}->insert($package, 1);
+        $self->{packages_trie}->insert($package, $count);
     }
+
+    return;
 } ## end sub load_trie
 
+sub start_indexing_function
+{
+    my ($self) = @_;
+
+    my $loop = IO::Async::Loop->new();
+
+    $self->{indexing_function} = IO::Async::Function->new(
+        min_workers => 0,
+        max_workers => 1,
+        code        => sub {
+            my ($self, @files) = @_;
+
+            # Lock the index file for this critical section of code
+            open my $fh, '>>', $self->{location} or die;
+            flock $fh, LOCK_EX;
+
+            my $index = $self->index(1); # 1 indicates that we should not try to lock again
+
+            unless (scalar @files)
+            {
+                @files = @{$self->get_all_perl_files()};
+                $self->_cleanup_old_files($index);
+                @files = grep { not exists $index->{files}{$_}{last_mtime} or not length $index->{files}{$_}{last_mtime} and $index->{files}{$_}{last_mtime} < stat($_)->mtime } @files;
+            } ## end unless (scalar @files)
+
+            my $total   = scalar @files;
+            my $current = 0;
+
+            $self->load_trie();
+
+            foreach my $file (@files)
+            {
+                $current++;
+
+                open my $fh, '<', $file or next;
+                my $text     = do { local $/; <$fh> };
+                my $document = PLS::Parser::Document->new(path => $file, text => \$text, no_cache => 1);
+                next unless (ref $document eq 'PLS::Parser::Document');
+
+                my $log_message = "Indexing $file";
+                $log_message .= " ($current/$total)" if (scalar @files > 1);
+                $log_message .= '...';
+
+                $self->log($log_message);
+
+                $self->update_subroutines($index, $document);
+                $self->update_packages($index, $document);
+            } ## end foreach my $file (@files)
+
+            $self->save($index);
+
+            flock $fh, LOCK_UN;
+
+            return [@{$self}{qw(cache last_mtime subs_trie packages_trie)}];
+        }
+    );
+
+    $loop->add($self->{indexing_function});
+
+    return;
+} ## end sub start_indexing_function
+
+sub start_cleanup_function
+{
+    my ($self) = @_;
+
+    my $loop = IO::Async::Loop->new();
+
+    $self->{cleanup_function} = IO::Async::Function->new(
+        min_workers => 0,
+        max_workers => 1,
+        code        => sub {
+            my ($self) = @_;
+
+            # Lock the index file for this critical section of code
+            open my $fh, '>>', $self->{location} or die;
+            flock $fh, LOCK_EX;
+
+            my $index = $self->index(1); # 1 indicates that we should not try to lock again
+            $self->_cleanup_old_files($index);
+            $self->save($index);
+
+            flock $fh, LOCK_UN;
+
+            return [@{$self}{qw(cache last_mtime)}];
+        }
+    );
+
+    $loop->add($self->{cleanup_function});
+
+    return;
+} ## end sub start_cleanup_function
+
 sub index_files
 {
     my ($self, @files) = @_;
 
-    async
-    {
-        my ($self, @files) = @_;
+    my $class     = ref $self;
+    my $self_copy = bless {%{$self}{qw(root location cache last_mtime subs_trie packages_trie)}}, $class;
 
-        my $lock  = $self->lock();
-        my $index = $self->index();
+    $self->{indexing_function}->call(
+        args      => [$self_copy, @files],
+        on_result => sub {
+            my ($result, $data) = @_;
 
-        unless (scalar @files)
-        {
-            @files = @{$self->get_all_perl_files()};
-            $self->_cleanup_old_files($index);
-            @files = grep { not exists $index->{files}{$_}{last_mtime} or not length $index->{files}{$_}{last_mtime} and $index->{files}{$_}{last_mtime} < stat($_)->mtime } @files;
-        } ## end unless (scalar @files)
+            return if ($result ne 'return');
+            @{$self}{qw(cache last_mtime subs_trie packages_trie)} = @{$data};
+        }
+    );
 
-        my $total   = scalar @files;
-        my $current = 0;
-
-        foreach my $file (@files)
-        {
-            $current++;
-
-            open my $fh, '<', $file or next;
-            my $text     = do { local $/; <$fh> };
-            my $document = PLS::Parser::Document->new(path => $file, text => \$text, no_cache => 1);
-            next unless (ref $document eq 'PLS::Parser::Document');
-
-            Coro::AnyEvent::sleep 0.01;
-
-            my $log_message = "Indexing $file";
-            $log_message .= " ($current/$total)" if (scalar @files > 1);
-            $log_message .= '...';
-
-            $self->log($log_message);
-
-            $self->update_subroutines($index, $document);
-            Coro::AnyEvent::sleep 0.01;
-            $self->update_packages($index, $document);
-            Coro::AnyEvent::sleep 0.01;
-        } ## end foreach my $file (@files)
-
-        $self->save($index);
-    } ## end async
-    $self, @files;
+    return;
 } ## end sub index_files
 
-sub lock
-{
-    return $indexing_semaphore->guard();
-}
-
 sub save
 {
     my ($self, $index) = @_;
@@ -133,23 +195,28 @@ sub save
     File::Path::make_path($parent_dir);
 
     Storable::nstore($index, $self->{location});
+
     $self->{cache}      = $index;
     $self->{last_mtime} = (stat $self->{location})->mtime;
+
+    return;
 } ## end sub save
 
 sub index
 {
-    my ($self) = @_;
+    my ($self, $no_lock) = @_;
 
     return {} unless -f $self->{location};
 
-    # Don't have two coroutines retrieving the index at the same time
-    my $guard = $self->{retrieve_index_semaphore}->guard();
     my $mtime = (stat $self->{location})->mtime;
     return $self->{cache} if ($mtime <= $self->{last_mtime});
 
     $self->{last_mtime} = $mtime;
-    $self->{cache}      = Storable::retrieve($self->{location});
+
+    open my $fh, '<', $self->{location} or die;
+    flock $fh, LOCK_SH unless $no_lock;
+    $self->{cache}      = Storable::fd_retrieve($fh);
+    flock $fh, LOCK_UN unless $no_lock;
 
     return $self->{cache};
 } ## end sub index
@@ -163,6 +230,7 @@ sub update_subroutines
 
     $self->cleanup_index($index, 'subs', $document->{path});
     $self->update_index($index, 'subs', $document->{path}, @$subroutines, @$constants);
+    return;
 } ## end sub update_subroutines
 
 sub update_packages
@@ -174,6 +242,7 @@ sub update_packages
     $self->cleanup_index($index, 'packages', $document->{path});
     return unless (ref $packages eq 'ARRAY');
     $self->update_index($index, 'packages', $document->{path}, @$packages);
+    return;
 } ## end sub update_packages
 
 sub cleanup_index
@@ -203,6 +272,8 @@ sub cleanup_index
     {
         $index->{files}{$file}{$type} = [];
     }
+
+    return;
 } ## end sub cleanup_index
 
 sub update_index
@@ -243,19 +314,28 @@ sub update_index
     } ## end foreach my $reference (@references...)
 
     $index->{files}{$file}{last_mtime} = $stat->mtime;
+
+    return;
 } ## end sub update_index
 
 sub cleanup_old_files
 {
     my ($self) = @_;
 
-    async
-    {
-        my $lock  = $self->lock();
-        my $index = $self->index();
-        $self->_cleanup_old_files($index);
-        $self->save($index);
-    } ## end async
+    my $class     = ref $self;
+    my $self_copy = bless {%{$self}{qw(root location cache last_mtime)}}, $class;
+
+    $self->{cleanup_function}->call(
+        args      => [$self_copy],
+        on_result => sub {
+            my ($result, $data) = @_;
+
+            return if ($result ne 'return');
+            @{$self}{qw(cache last_mtime)} = @{$data};
+        }
+    );
+
+    return;
 } ## end sub cleanup_old_files
 
 sub _cleanup_old_files
@@ -304,10 +384,12 @@ sub _cleanup_old_files
             @{$index->{$type}{$ref}} = grep { -e $_->{file} } @{$index->{$type}{$ref}};
             my $count_after = scalar @{$index->{$type}{$ref}};
             $refs_cleaned++ if ($count_after < $count_before);
-        }
+        } ## end foreach my $ref (keys %{$index...})
 
         $self->log("Cleaning up $type references from index...") if ($refs_cleaned);
-    }
+    } ## end foreach my $type (keys %{$index...})
+
+    return;
 } ## end sub _cleanup_old_files
 
 sub find_package_subroutine
@@ -509,6 +591,8 @@ sub log
     my $time = Time::Piece->new;
     $time = $time->ymd . ' ' . $time->hms;
     print {\*STDERR} "[$time] $message\n";
+
+    return;
 } ## end sub log
 
 1;
